/*
 * Copyright (c) 2010-2015 A8CAS developers (see AUTHORS)
 *
 * This file is part of the A8CAS project which allows to manipulate tape
 * images for Atari 8-bit computers.
 *
 * A8CAS is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * A8CAS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with A8CAS; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 */
#include "pwm_demod.h"
#include <stdio.h>
#include <math.h>

void PWM_DEMOD_init(PWM_DEMOD_t *demod)
{
	demod->prev_sample = 0;
	demod->prev_up = 0;
	demod->last_width = demod->lookaround;
}

int PWM_DEMOD_advance(PWM_DEMOD_t *demod, short int sample)
{
	static int counter = 0;
	int min_counter;
	int max_counter;
	static int diff = 0;
	int noise_level = (demod->lookaround == 0 ? 0 : ceil(sqrt((double)demod->amplitude * demod->amplitude / pow(10,(demod->signal_to_noise / 10)))));
	static int edgecounter = 0;

/*	printf("amplitude=%d, noise_level=%d, tolerance=%d, lookaround=%d\n",demod->amplitude,noise_level,demod->tolerance,demod->lookaround);*/

	min_counter = demod->lookaround;
	max_counter = min_counter * 2;

	counter++;

	if ((demod->lookaround != 0) &&
	   ((demod->amplitude == 0) ||
	   (counter > 15 * demod->samplerate / 48000 &&  demod->amplitude < demod->silence_noise)))
		return demod->prev_up;

	if (demod->prev_up) {
		if (demod->prev_sample > sample + noise_level) {
			diff += demod->prev_sample - sample;
			edgecounter++;
		} else {
			diff = 0;
			edgecounter = 0;
		}
/*		printf("prev_UP, prev=%d, curr=%d, diff=%d, last_width=%d, counter=%d\n",demod->prev_sample,sample,diff,demod->last_width,counter);*/
	}
	else {
		if (sample > demod->prev_sample + noise_level) {
			diff += sample - demod->prev_sample;
			edgecounter++;
		} else {
			diff = 0;
			edgecounter = 0;
		}
/*		printf("prev_DOWN, prev=%d, curr=%d, diff=%d, last_width=%d, counter=%d\n",demod->prev_sample,sample,diff,demod->last_width,counter);*/
	}
	demod->prev_sample = sample;
	if (diff + noise_level >= demod->tolerance /* steep slope, sharp edge (or tolerance manual)*/
	|| (demod->lookaround != 0 && edgecounter >= demod->last_width/4)) { /* gentle slope, smooth edge (tolerance auto)*/
		if (counter < demod->lookaround/4) { /* very poor "lowpass filter" */
			return demod->prev_up;
		}
		demod->prev_up = !demod->prev_up;
/*		demod->last_width = counter + 2;*/
		demod->last_width = (counter < min_counter) ? min_counter + 2 :
					counter > max_counter ? max_counter :
/*					counter;*/
					(max_counter - counter) < (counter - min_counter) ? max_counter:
					min_counter + 2;
		edgecounter = 0;
		counter = 0;
		diff=0;
	}
	if (counter > 200 * demod->samplerate / 48000) /* silence */
		counter = 0;
	return demod->prev_up;
}
